#!/usr/bin/python
# -*- encoding: utf-8; py-indent-offset: 4 -*-
# +------------------------------------------------------------------+
# |             ____ _               _        __  __ _  __           |
# |            / ___| |__   ___  ___| | __   |  \/  | |/ /           |
# |           | |   | '_ \ / _ \/ __| |/ /   | |\/| | ' /            |
# |           | |___| | | |  __/ (__|   <    | |  | | . \            |
# |            \____|_| |_|\___|\___|_|\_\___|_|  |_|_|\_\           |
# |                                                                  |
# | Copyright Mathias Kettner 2014             mk@mathias-kettner.de |
# +------------------------------------------------------------------+
#
# This file is part of Check_MK.
# The official homepage is at http://mathias-kettner.de/check_mk.
#
# check_mk is free software;  you can redistribute it and/or modify it
# under the  terms of the  GNU General Public License  as published by
# the Free Software Foundation in version 2.  check_mk is  distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY;  with-
# out even the implied warranty of  MERCHANTABILITY  or  FITNESS FOR A
# PARTICULAR PURPOSE. See the  GNU General Public License for more de-
# tails. You should have  received  a copy of the  GNU  General Public
# License along with GNU Make; see the file  COPYING.  If  not,  write
# to the Free Software Foundation, Inc., 51 Franklin St,  Fifth Floor,
# Boston, MA 02110-1301 USA.

import os
import ast
import pprint
import sys
import tarfile
import time
import subprocess
import json
from cStringIO import StringIO
from typing import NamedTuple

import cmk.ec.export
import cmk.utils.log
import cmk.utils.paths
import cmk.utils.tty as tty
import cmk.utils.werks
import cmk.utils.debug
import cmk_base.utils

logger = cmk.utils.log.get_logger(__name__)
pac_ext = ".mkp"


# TODO: Subclass MKGeneralException()?
class PackageException(Exception):
    def __init__(self, reason):
        self.reason = reason
        super(PackageException, self).__init__(reason)

    def __str__(self):
        return self.reason


pac_dir = cmk.utils.paths.omd_root + "/var/check_mk/packages/"

# TODO: OMD: Pack this path and remote this makedirs call
try:
    os.makedirs(pac_dir)
except:
    pass

# order matters! See function _get_permissions
PERM_MAP = (
    (cmk.utils.paths.checks_dir, 0644),
    (cmk.utils.paths.local_checks_dir, 0644),
    (cmk.utils.paths.notifications_dir, 0755),
    (cmk.utils.paths.local_notifications_dir, 0755),
    (cmk.utils.paths.inventory_dir, 0644),
    (cmk.utils.paths.local_inventory_dir, 0644),
    (cmk.utils.paths.check_manpages_dir, 0644),
    (cmk.utils.paths.local_check_manpages_dir, 0644),
    (cmk.utils.paths.agents_dir, 0755),
    (cmk.utils.paths.local_agents_dir, 0755),
    (cmk.utils.paths.web_dir, 0644),
    (cmk.utils.paths.local_web_dir, 0644),
    (cmk.utils.paths.pnp_templates_dir, 0644),
    (cmk.utils.paths.local_pnp_templates_dir, 0644),
    (cmk.utils.paths.doc_dir, 0644),
    (cmk.utils.paths.local_doc_dir, 0644),
    (cmk.utils.paths.locale_dir, 0644),
    (cmk.utils.paths.local_locale_dir, 0644),
    (cmk.utils.paths.local_bin_dir, 0755),
    (os.path.join(cmk.utils.paths.local_lib_dir, "nagios", "plugins"), 0755),
    (cmk.utils.paths.local_lib_dir, 0644),
    (cmk.utils.paths.local_mib_dir, 0644),
    (os.path.join(cmk.utils.paths.share_dir, "alert_handlers"), 0755),
    (os.path.join(cmk.utils.paths.local_share_dir, "alert_handlers"), 0755),
    (str(cmk.ec.export.mkp_rule_pack_dir()), 0644),
)


def _get_permissions(path):
    """Determine permissions by the first matching beginning of 'path'"""
    for path_begin, perm in PERM_MAP:
        if path.startswith(path_begin):
            return perm
    raise PackageException("could not determine permissions for %r" % path)


PackagePart = NamedTuple("PackagePart", [
    ("ident", str),
    ("title", str),
    ("path", str),
])

_package_parts = [
    PackagePart("checks", "Checks", cmk.utils.paths.local_checks_dir),
    PackagePart("notifications", "Notification scripts", cmk.utils.paths.local_notifications_dir),
    PackagePart("inventory", "Inventory plugins", cmk.utils.paths.local_inventory_dir),
    PackagePart("checkman", "Checks' man pages", cmk.utils.paths.local_check_manpages_dir),
    PackagePart("agents", "Agents", cmk.utils.paths.local_agents_dir),
    PackagePart("web", "Multisite extensions", cmk.utils.paths.local_web_dir),
    PackagePart("pnp-templates", "PNP4Nagios templates", cmk.utils.paths.local_pnp_templates_dir),
    PackagePart("doc", "Documentation files", cmk.utils.paths.local_doc_dir),
    PackagePart("locales", "Localizations", cmk.utils.paths.local_locale_dir),
    PackagePart("bin", "Binaries", cmk.utils.paths.local_bin_dir),
    PackagePart("lib", "Libraries", cmk.utils.paths.local_lib_dir),
    PackagePart("mibs", "SNMP MIBs", cmk.utils.paths.local_mib_dir),
    PackagePart("alert_handlers", "Alert handlers",
                cmk.utils.paths.local_share_dir + "/alert_handlers"),
]

config_parts = [
    PackagePart("ec_rule_packs", "Event Console rule packs",
                str(cmk.ec.export.mkp_rule_pack_dir())),
]

package_ignored_files = {
    "lib": ["nagios/plugins/README.txt"],
}


def get_package_parts():
    return _package_parts


def packaging_usage():
    sys.stdout.write("""Usage: check_mk [-v] -P|--package COMMAND [ARGS]

Available commands are:
   create NAME      ...  Collect unpackaged files into new package NAME
   pack NAME        ...  Create package file from installed package
   release NAME     ...  Drop installed package NAME, release packaged files
   find             ...  Find and display unpackaged files
   list             ...  List all installed packages
   list NAME        ...  List files of installed package
   list PACK.mkp    ...  List files of uninstalled package file
   show NAME        ...  Show information about installed package
   show PACK.mkp    ...  Show information about uninstalled package file
   install PACK.mkp ...  Install or update package from file PACK.mkp
   remove NAME      ...  Uninstall package NAME

   -v  enables verbose output

Package files are located in %s.
""" % pac_dir)


def do_packaging(args):
    if len(args) == 0:
        packaging_usage()
        sys.exit(1)
    command = args[0]
    args = args[1:]

    commands = {
        "create": package_create,
        "release": package_release,
        "list": package_list,
        "find": package_find,
        "show": package_info,
        "pack": package_pack,
        "remove": package_remove,
        "install": package_install,
    }
    f = commands.get(command)
    if f:
        try:
            f(args)
        except PackageException as e:
            logger.error("%s" % e)
            sys.exit(1)
    else:
        allc = commands.keys()
        allc.sort()
        allc = [tty.bold + c + tty.normal for c in allc]
        logger.error("Invalid packaging command. Allowed are: %s and %s.", ", ".join(allc[:-1]),
                     allc[-1])
        sys.exit(1)


def package_list(args):
    if len(args) > 0:
        for name in args:
            show_package_contents(name)
    else:
        if logger.is_verbose():
            table = []
            for pacname in all_package_names():
                package = read_package_info(pacname)
                table.append((pacname, package["title"], package["num_files"]))
            tty.print_table(["Name", "Title", "Files"], [tty.bold, "", ""], table)
        else:
            for pacname in all_package_names():
                sys.stdout.write("%s\n" % pacname)


def package_info(args):
    if len(args) == 0:
        raise PackageException("Usage: check_mk -P show NAME|PACKAGE.mkp")
    for name in args:
        show_package_info(name)


def show_package_contents(name):
    show_package(name, False)


def show_package_info(name):
    show_package(name, True)


def show_package(name, show_info=False):
    try:
        if name.endswith(pac_ext):
            tar = tarfile.open(name, "r:gz")
            info = tar.extractfile("info")
            package = parse_package_info(info.read())
        else:
            package = read_package_info(name)
            if not package:
                raise PackageException("No such package %s." % name)
            if show_info:
                sys.stdout.write("Package file:                  %s%s\n" % (pac_dir, name))
    except PackageException:
        raise
    except Exception as e:
        raise PackageException("Cannot open package %s: %s" % (name, e))

    if show_info:
        sys.stdout.write("Name:                          %s\n" % package["name"])
        sys.stdout.write("Version:                       %s\n" % package["version"])
        sys.stdout.write("Packaged on Checkmk Version:   %s\n" % package["version.packaged"])
        sys.stdout.write("Required Checkmk Version:      %s\n" % package["version.min_required"])
        valid_until_text = package["version.usable_until"] or "No version limitation"
        sys.stdout.write("Valid until Checkmk version:   %s\n" % valid_until_text)
        sys.stdout.write("Title:                         %s\n" % package["title"])
        sys.stdout.write("Author:                        %s\n" % package["author"])
        sys.stdout.write("Download-URL:                  %s\n" % package["download_url"])
        sys.stdout.write("Files:                         %s\n" % \
                " ".join([ "%s(%d)" % (part, len(fs)) for part, fs in package["files"].items() ]))
        sys.stdout.write("Description:\n  %s\n" % package["description"])
    else:
        if logger.is_verbose():
            sys.stdout.write("Files in package %s:\n" % name)
            for part in get_package_parts():
                files = package["files"].get(part.ident, [])
                if len(files) > 0:
                    sys.stdout.write("  %s%s%s:\n" % (tty.bold, part.title, tty.normal))
                    for f in files:
                        sys.stdout.write("    %s\n" % f)
        else:
            for part in get_package_parts():
                for fn in package["files"].get(part.ident, []):
                    sys.stdout.write(part.path + "/" + fn + "\n")


def package_create(args):
    if len(args) != 1:
        raise PackageException("Usage: check_mk -P create NAME")

    pacname = args[0]
    if read_package_info(pacname):
        raise PackageException("Package %s already existing." % pacname)

    logger.verbose("Creating new package %s...", pacname)
    filelists = {}
    package = {
        "title": "Title of %s" % pacname,
        "name": pacname,
        "description": "Please add a description here",
        "version": "1.0",
        "version.packaged": cmk.__version__,
        "version.min_required": cmk.__version__,
        "version.usable_until": None,
        "author": "Add your name here",
        "download_url": "http://example.com/%s/" % pacname,
        "files": filelists
    }
    num_files = 0
    for part in get_package_parts():
        files = unpackaged_files_in_dir(part.ident, part.path)
        filelists[part.ident] = files
        num_files += len(files)
        if len(files) > 0:
            logger.verbose("  %s%s%s:", tty.bold, part.title, tty.normal)
            for f in files:
                logger.verbose("    %s", f)

    write_package_info(package)
    logger.verbose("New package %s created with %d files.", pacname, num_files)
    logger.verbose("Please edit package details in %s%s%s", tty.bold, pac_dir + pacname, tty.normal)


def package_find(_no_args):
    first = True
    for part in get_package_parts() + config_parts:
        files = unpackaged_files_in_dir(part.ident, part.path)
        if len(files) > 0:
            if first:
                logger.verbose("Unpackaged files:")
                first = False

            logger.verbose("  %s%s%s:", tty.bold, part.title, tty.normal)
            for f in files:
                if logger.is_verbose():
                    logger.verbose("    %s", f)
                else:
                    logger.info("%s/%s", part.path, f)

    if first:
        logger.verbose("No unpackaged files found.")


def release_package(pacname):
    if not pacname or not package_exists(pacname):
        raise PackageException("Package %s not installed or corrupt." % pacname)

    package = read_package_info(pacname)
    logger.verbose("Releasing files of package %s into freedom...", pacname)
    for part in get_package_parts() + config_parts:
        filenames = package["files"].get(part.ident, [])
        if len(filenames) > 0:
            logger.verbose("  %s%s%s:", tty.bold, part.title, tty.normal)
            for f in filenames:
                logger.verbose("    %s", f)
            if part.ident == 'ec_rule_packs':
                cmk.ec.export.release_packaged_rule_packs(filenames)
    remove_package_info(pacname)


def package_release(args):
    if len(args) != 1:
        raise PackageException("Usage: check_mk -P release NAME")
    pacname = args[0]
    release_package(pacname)


def package_exists(pacname):
    pacpath = pac_dir + pacname
    return os.path.exists(pacpath)


def package_pack(args):
    if len(args) != 1:
        raise PackageException("Usage: check_mk -P pack NAME")

    # Make sure, user is not in data directories of Check_MK
    abs_curdir = os.path.abspath(os.curdir)
    for directory in [cmk.utils.paths.var_dir
                     ] + [p.path for p in get_package_parts() + config_parts]:
        if abs_curdir == directory or abs_curdir.startswith(directory + "/"):
            raise PackageException(
                "You are in %s!\n"
                "Please leave the directories of Check_MK before creating\n"
                "a packet file. Foreign files lying around here will mix up things." % abs_curdir)

    pacname = args[0]
    package = read_package_info(pacname)
    if not package:
        raise PackageException("Package %s not existing or corrupt." % pacname)
    tarfilename = "%s-%s%s" % (pacname, package["version"], pac_ext)
    logger.verbose("Packing %s into %s...", pacname, tarfilename)
    create_mkp_file(package, file_name=tarfilename)
    logger.verbose("Successfully created %s", tarfilename)


def create_mkp_file(package, file_name=None, file_object=None):
    package["version.packaged"] = cmk.__version__

    def create_tar_info(filename, size):
        info = tarfile.TarInfo()
        info.mtime = int(time.time())
        info.uid = 0
        info.gid = 0
        info.size = size
        info.mode = 0644
        info.type = tarfile.REGTYPE
        info.name = filename
        return info

    tar = tarfile.open(name=file_name, fileobj=file_object, mode="w:gz")

    # add the regular info file (Python format)
    info_file = StringIO(pprint.pformat(package))
    info = create_tar_info("info", len(info_file.getvalue()))
    tar.addfile(info, info_file)

    # add the info file a second time (JSON format) for external tools
    info_file = StringIO(json.dumps(package))
    info = create_tar_info("info.json", len(info_file.getvalue()))
    tar.addfile(info, info_file)

    # Now pack the actual files into sub tars
    for part in get_package_parts() + config_parts:
        filenames = package["files"].get(part.ident, [])
        if len(filenames) > 0:
            logger.verbose("  %s%s%s:", tty.bold, part.title, tty.normal)
            for f in filenames:
                logger.verbose("    %s", f)
            subtarname = part.ident + ".tar"
            subdata = subprocess.check_output(
                ["tar", "cf", "-", "--dereference", "--force-local", "-C", part.path] + filenames)
            info = create_tar_info(subtarname, len(subdata))
            tar.addfile(info, StringIO(subdata))
    tar.close()


def package_remove(args):
    if len(args) != 1:
        raise PackageException("Usage: check_mk -P remove NAME")
    pacname = args[0]
    package = read_package_info(pacname)
    if not package:
        raise PackageException("No such package %s." % pacname)

    logger.verbose("Removing package %s...", pacname)
    remove_package(package)
    logger.verbose("Successfully removed package %s.", pacname)


def remove_package(package):
    for part in get_package_parts() + config_parts:
        filenames = package["files"].get(part.ident, [])
        if len(filenames) > 0:
            logger.verbose("  %s%s%s", tty.bold, part.title, tty.normal)
            for fn in filenames:
                logger.verbose("    %s", fn)
                try:
                    path = part.path + "/" + fn
                    if part.ident == 'ec_rule_packs':
                        cmk.ec.export.remove_packaged_rule_packs(filenames)
                    else:
                        os.remove(path)
                except Exception as e:
                    if cmk.utils.debug.enabled():
                        raise
                    raise Exception("Cannot remove %s: %s\n" % (path, e))

    os.remove(pac_dir + package["name"])


def create_package(pkg_info):
    pacname = pkg_info["name"]
    if package_exists(pacname):
        raise PackageException("Packet already exists.")

    validate_package_files(pacname, pkg_info["files"])
    write_package_info(pkg_info)


def edit_package(pacname, new_package_info):
    if not package_exists(pacname):
        raise PackageException("No such package")

    # Renaming: check for collision
    if pacname != new_package_info["name"]:
        if package_exists(new_package_info["name"]):
            raise PackageException(
                "Cannot rename package: a package with that name already exists.")

    validate_package_files(pacname, new_package_info["files"])

    remove_package_info(pacname)
    write_package_info(new_package_info)


def install_optional_package(package_file_name):
    if package_file_name not in [p.name.decode("utf-8") for p in get_optional_package_paths()]:
        raise PackageException("Optional package %s does not exist" % package_file_name)

    return install_package(
        "%s" % cmk.utils.paths.optional_packages_dir.joinpath(package_file_name.encode("utf-8")))


# Packaged files must either be unpackaged or already
# belong to that package
def validate_package_files(pacname, files):
    packages = {}
    for package_name in all_package_names():
        packages[package_name] = read_package_info(package_name)

    for part in get_package_parts():
        validate_package_files_part(packages, pacname, part.ident, part.path,
                                    files.get(part.ident, []))


def validate_package_files_part(packages, pacname, part, directory, rel_paths):
    for rel_path in rel_paths:
        path = os.path.join(directory, rel_path)
        if not os.path.exists(path):
            raise PackageException("File %s does not exist." % path)

        for other_pacname, other_package_info in packages.items():
            for other_rel_path in other_package_info["files"].get(part, []):
                if other_rel_path == rel_path and other_pacname != pacname:
                    raise PackageException("File %s does already belong to package %s" %
                                           (path, other_pacname))


def package_install(args):
    if len(args) != 1:
        raise PackageException("Usage: check_mk -P install NAME")
    path = args[0]
    if not os.path.exists(path):
        raise PackageException("No such file %s." % path)

    return install_package(file_name=path)


def install_package(file_name=None, file_object=None):
    tar = tarfile.open(name=file_name, fileobj=file_object, mode="r:gz")
    package = parse_package_info(tar.extractfile("info").read())

    verify_check_mk_version(package)

    pacname = package["name"]
    old_package = read_package_info(pacname)
    if old_package:
        logger.verbose("Updating %s from version %s to %s.", pacname, old_package["version"],
                       package["version"])
        update = True
    else:
        logger.verbose("Installing %s version %s.", pacname, package["version"])
        update = False

    # Before installing check for conflicts
    keep_files = {}
    for part in get_package_parts() + config_parts:
        packaged = packaged_files_in_dir(part.ident)
        keep = []
        keep_files[part.ident] = keep

        if update:
            old_files = old_package["files"].get(part.ident, [])

        for fn in package["files"].get(part.ident, []):
            path = os.path.join(part.path, fn)
            if update and fn in old_files:
                keep.append(fn)
            elif fn in packaged:
                raise PackageException("File conflict: %s is part of another package." % path)
            elif os.path.exists(path):
                raise PackageException("File conflict: %s already existing." % path)

    # Now install files, but only unpack files explicitely listed
    for part in get_package_parts() + config_parts:
        filenames = package["files"].get(part.ident, [])
        if len(filenames) > 0:
            logger.verbose("  %s%s%s:", tty.bold, part.title, tty.normal)
            for fn in filenames:
                logger.verbose("    %s", fn)

            # make sure target directory exists
            if not os.path.exists(part.path):
                logger.verbose("    Creating directory %s", part.path)
                os.makedirs(part.path)

            tarsource = tar.extractfile(part.ident + ".tar")

            # Important: Do not preserve the tared timestamp. Checkmk needs to know when the files
            # been installed for cache invalidation.
            tardest = subprocess.Popen(["tar", "xf", "-", "--touch", "-C", part.path] + filenames,
                                       stdin=subprocess.PIPE,
                                       shell=False,
                                       close_fds=True)
            while True:
                data = tarsource.read(4096)
                if not data:
                    break
                tardest.stdin.write(data)

            tardest.stdin.close()
            tardest.wait()

            # Fix permissions of extracted files
            for filename in filenames:
                path = os.path.join(part.path, filename)
                desired_perm = _get_permissions(path)
                has_perm = os.stat(path).st_mode & 07777
                if has_perm != desired_perm:
                    logger.verbose("    Fixing permissions of %s: %04o -> %04o", path, has_perm,
                                   desired_perm)
                    os.chmod(path, desired_perm)

            if part.ident == 'ec_rule_packs':
                cmk.ec.export.add_rule_pack_proxies(filenames)

    # In case of an update remove files from old_package not present in new one
    if update:
        for part in get_package_parts() + config_parts:
            filenames = old_package["files"].get(part.ident, [])
            keep = keep_files.get(part.ident, [])
            for fn in filenames:
                if fn not in keep:
                    path = os.path.join(part.path, fn)
                    logger.verbose("Removing outdated file %s.", path)
                    try:
                        os.remove(path)
                    except Exception as e:
                        logger.error("Error removing %s: %s", path, e)

            if part.ident == 'ec_rule_packs':
                to_remove = [fn for fn in filenames if fn not in keep]
                cmk.ec.export.remove_packaged_rule_packs(to_remove, delete_export=False)

    # Last but not least install package file
    write_package_info(package)
    return package


# Checks whether or not the minimum required Check_MK version is older than the
# current Check_MK version. Raises an exception if not. When the Check_MK version
# can not be parsed or is a daily build, the check is simply passing without error.
def verify_check_mk_version(package):
    min_version = package["version.min_required"]
    cmk_version = cmk.__version__

    if cmk_base.utils.is_daily_build_version(min_version):
        min_branch = cmk_base.utils.branch_of_daily_build(min_version)
        if min_branch == "master":
            return  # can not check exact version
        else:
            # use the branch name (e.g. 1.2.8 as min version)
            min_version = min_branch

    if cmk_base.utils.is_daily_build_version(cmk_version):
        branch = cmk_base.utils.branch_of_daily_build(cmk_version)
        if branch == "master":
            return  # can not check exact version
        else:
            # use the branch name (e.g. 1.2.8 as min version)
            cmk_version = branch

    compatible = True
    try:
        compatible = cmk.utils.werks.parse_check_mk_version(min_version) \
                        <= cmk.utils.werks.parse_check_mk_version(cmk_version)
    except:
        # Be compatible: When a version can not be parsed, then skip this check
        if cmk.utils.debug.enabled():
            raise
        return

    if not compatible:
        raise PackageException("The package requires Check_MK version %s, "
                               "but you have %s installed." % (min_version, cmk_version))


def files_in_dir(part, directory, prefix=""):
    if directory is None or not os.path.exists(directory):
        return []

    # Handle case where one part-directory lies below another
    taboo_dirs = [p.path for p in get_package_parts() + config_parts if p.ident != part]
    if directory in taboo_dirs:
        return []

    result = []
    files = os.listdir(directory)
    for f in files:
        if f in ['.', '..'] or f.startswith('.') or f.endswith('~') or f.endswith(".pyc"):
            continue

        ignored = package_ignored_files.get(part, [])
        if prefix + f in ignored:
            continue

        path = directory + "/" + f
        if os.path.isdir(path):
            result += files_in_dir(part, path, prefix + f + "/")
        else:
            result.append(prefix + f)
    result.sort()
    return result


def get_optional_package_infos():
    optional = {}
    for pkg_path in get_optional_package_paths():
        with pkg_path.open("rb") as pkg:
            pkg_info = get_package_info_from_package(pkg)
            optional[pkg_path.name.decode("utf-8")] = pkg_info

    return optional


def get_package_info_from_package(file_object):
    tar = tarfile.open(fileobj=file_object, mode="r:gz")
    package_info_file = tar.extractfile("info")
    if package_info_file is None:
        raise PackageException("Failed to open package info file")
    return parse_package_info(package_info_file.read())


def get_optional_package_paths():
    if not cmk.utils.paths.optional_packages_dir.exists():
        return []
    return list(cmk.utils.paths.optional_packages_dir.iterdir())


def unpackaged_files():
    unpackaged = {}
    for part in get_package_parts() + config_parts:
        unpackaged[part.ident] = unpackaged_files_in_dir(part.ident, part.path)
    return unpackaged


def package_part_info():
    part_info = {}
    for part in get_package_parts() + config_parts:
        try:
            files = os.listdir(part.path)
        except OSError:
            files = []

        part_info[part.ident] = {
            "title": part.title,
            "permissions": map(_get_permissions, [os.path.join(part.path, f) for f in files]),
            "path": part.path,
            "files": files,
        }

    return part_info


def unpackaged_files_in_dir(part, directory):
    packaged = set(packaged_files_in_dir(part))
    return [f for f in files_in_dir(part, directory) if f not in packaged]


def packaged_files_in_dir(part):
    result = []
    for pacname in all_package_names():
        package = read_package_info(pacname)
        if package:
            result += package["files"].get(part, [])
    return result


def read_package_info(pacname):
    try:
        package = parse_package_info(file(pac_dir + pacname).read())
        package["name"] = pacname  # do not trust package content
        num_files = sum([len(fl) for fl in package["files"].values()])
        package["num_files"] = num_files
        return package
    except IOError:
        return None
    except Exception:
        logger.verbose("Ignoring invalid package file '%s%s'. Please remove it from %s!", pac_dir,
                       pacname, pac_dir)
        return None


def write_package_info(package):
    file(pac_dir + package["name"], "w").write(pprint.pformat(package) + "\n")


def remove_package_info(pacname):
    os.remove(pac_dir + pacname)


def all_package_names():
    return sorted([p for p in os.listdir(pac_dir) if p not in ['.', '..']])


def parse_package_info(python_string):
    pkg_info = ast.literal_eval(python_string)
    pkg_info.setdefault("version.usable_until", None)
    return pkg_info
